
/**
 *  I originally created this to be a small, single script solution based on my library changes.
 *  however in the DLL i was free to modify a lot of things. Here, forced to work with only public 
 *  elements, and limited ui display interception, it's grown a little bit complicated.
 *  At some point I'll probably break this into multiple components or otherwise encapsulate concerns.
 *  in the meantime here's a big script.
 *  
 *  if it's any consolation it's 90% comments.
 * */


namespace GossamerVR.MorphFilterPlus
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Text.RegularExpressions;
    using UnityEngine;
    using UnityEngine.UI;
    using SimpleJSON;


    /// <summary>
    /// Upgrades person morph search bars to have AND, NOT, OR, STARTSWITH and other filter controls.
    /// designed to be a session plugin
    /// </summary>
    public class MorphFilterPlus : MVRScript
    {
        // strings are dangerous. It's easy to misspell a string, and then 
        // you're spending hours trying to figure out what went wrong.
        // so we put our strings all in one place, defined in advance, up here.
        // By making them fields and constants instead of hand-written strings, the compiler
        // will alert us if we misspell anything when we use them, AND we
        // only have to correct or change the string in a single place.
        ////////////////////// STRINGS ///////////////////////////////
        #region strings

        // all transform paths are relative to an atom transform.

        private const string k_morphUIPathFemale = "MorphAutoUIFemale";
        private const string k_morphUIPathMale = "MorphAutoUIMale";

        /// <summary>
        /// the transform path to the child with an AtomUI component, 
        /// for which we will be modifying the SaveAppearance button, 
        /// so that we can respond before the save.
        /// </summary>
        private const string k_personUIPath = "UI/UIPlaceHolderModel/UIModel";

        private const string k_personUIContentPath = k_personUIPath + "/Canvas/Panel/Content";

        // these are the transform paths to the filter input we're going to change.
        // there are two UIs (male & female) and an alt for each, because 
        // when you first select the person, VAM renames some UI bits-- thus, the path changes.
        // when we go to find the field we need to account for both paths of both UIs,
        // because we don't know if you've already selected the person or not.
        private const string k_inputFieldPathFemale = k_personUIContentPath + "/Female Morphs/FilterInputField";
        private const string k_inputFieldPathFemaleAlt = k_personUIContentPath + "/TabMorphsFemale/FilterInputField";        
        private const string k_inputFieldPathMale = k_personUIContentPath + "/Male Morphs/FilterInputField";
        private const string k_inputFieldPathMaleAlt = k_personUIContentPath + "/TabMorphsMale/FilterInputField";
        

        private const string k_personGeomPath = "rescale2/geometry";        

        /// <summary>
        /// path that we parent our appearance listener to. Adding a child to
        /// be considered 'part of the appearance' is the only way i've found so far
        /// to actually listen for the save of the appearance and react. 
        /// </summary>        
        private const string k_listenerParentPath = k_personGeomPath + "/FemaleCharacters";


        /// <summary>
        /// the array of morph banks collectively storing all the character morphs.
        /// we store them as an array so we can foreach() them easily. Their order
        /// is important because we'll use the MorphBankPath enum as indices.
        /// </summary>
        private static readonly string[] morphBankPaths = new[]
        {
            k_personGeomPath+ "/MorphBanks/FemaleCharacterMorphBank(Clone)",
            k_personGeomPath+ "/MorphBanks/FemaleGenMorphBank(Clone)",
            k_personGeomPath+ "/MorphBanks/MaleCharacterMorphBank(Clone)",
            k_personGeomPath+ "/MorphBanks/MaleGenMorphBank(Clone)",
            k_personGeomPath+ "/MorphBanks/MaleGen2MorphBank(Clone)",
        };

        private static class AtomType
        {
            public const string Person = "Person";
            public const string SubScene = "SubScene";
        }

        // while most of the time our bank paths are iterated over,
        // we are going to want to be able to access them specifically sometimes.
        // Rather than memorizing what each int index means, we use constants.
        private static class BankIndex
        {
            public const int FemaleChar = 0;
            public const int FemaleGen = 1;
            public const int MaleChar = 2;
            public const int MaleGen = 3;
            public const int MaleGen2 = 4;
        }
        // this is where I would be using an enum.
        //
        //                               IF I HAD ONE


        #endregion strings
        //////////////////////////////////////////////////////////////

        /// <summary>
        /// to let people know it's working we'll turn the input fields blue.
        /// </summary>
        private static readonly Color mintyFreshColor = new Color(0.85f, 0.95f, 1f, 1f);


        /// <summary>
        /// The UIDs of morphs VAM has already marked as invisible before we start.
        /// We store these to ensure we never set them visible.
        /// </summary>        
        private static HashSet<string> ignoredMorphs = null;


        /// <summary>        
        /// when true, the filter will behave as normal: applying every time
        /// the input field changes. When it's false, the filter only gets applied
        /// when the user clicks away from the field or presses enter.
        /// We show this in the plugin's UI
        /// </summary>
        private JSONStorableBool liveUpdateBoolJSON;

        /// <summary>
        /// when true, spaces will be converted to '+' symbol as a QOL option.
        /// </summary>
        private JSONStorableBool convertSpacesBoolJSON;

        /// <summary>
        /// allows the script to log more.
        /// </summary>
        private JSONStorableBool debugBoolJSON;

        /// <summary>
        /// the number of characters in the input field required before 
        /// the filter will begin to apply if live update is true.
        /// </summary>
        private JSONStorableFloat minCharacterFloatJSON;


        private bool isDebugMode => this.debugBoolJSON != null && this.debugBoolJSON.val;

        /// <summary>
        /// A table of the atoms we have "upgraded" and their upgrades,
        /// which store references to modifications.
        /// </summary>      
        private Dictionary<Atom, UpgradeProfile> upgradeLookup = new Dictionary<Atom, UpgradeProfile>();

        /// <summary>
        /// ongoing coroutine tasks to move through in update
        /// because I prefer having direct control instead of using StartCoroutine
        /// </summary>
        private List<IEnumerator> tasks = new List<IEnumerator>();

        private bool isInitialized = false;
        private bool isUIInitialized = false;
        private bool isSubscribed = false;
        
        /// <summary>
        /// When this string is populated, it means something went wrong.
        /// it will be logged and will tell the plugin to shut down.
        /// it allows us to gracefully disable when there's an error,
        /// and not spam the user
        /// </summary>
        private string generalError = "";

        private bool hasError => this.everythingIsTerrible || !string.IsNullOrEmpty(generalError);

        /// <summary>
        /// This flag is set when we self-disable due to critical errors,        
        /// it allows us to sleep, essentially, until reset.
        /// </summary>
        private bool everythingIsTerrible = false;


        /// <summary>
        /// We apply filters as a coroutine when a inputfield is submitted,
        /// to allow the clear button a chance to clear it. We also use this
        /// right after a filter is applied, to resync morphs after a few frames.
        /// </summary>
        private IEnumerator submissionTask = null;


        /// <summary>
        /// Re-subscribes to scene events if we are re-activating post-Initialization,
        /// and also scans for any new people.
        /// </summary>
        private void OnEnable()
        {
            // reset errors to try again.
            this.everythingIsTerrible = false;
            this.generalError = "";

            // if re-enabling after init: re-subscribe and re-scan.
            if (this.isInitialized && !this.isSubscribed)
            {
                this.Subscribe();
                this.ScanForUpgradableAtoms();
            }
        }

        /// <summary>
        /// Unsubscribes from scene events and reverts any upgrades to people we had.
        /// called automatically by Monobehaviour API when disabled.
        /// </summary>
        private void OnDisable()
        {
            this.LogIf(this.isDebugMode, "MorphFilterPlus Disabling");
            this.Unsubscribe();
            foreach (var keyValuePair in this.upgradeLookup)
            {
                this.LogIf(this.isDebugMode, "MorphFilterPlus reverting " + (keyValuePair.Key != null ? keyValuePair.Key.name : "NULL"));
                keyValuePair.Value?.Revert();
            }
            this.upgradeLookup.Clear();
            this.tasks.Clear();
            MorphFilterPlus.ignoredMorphs = null;
        }

        /// <summary>
        /// Initializes by subscribing to important scene events and 
        /// scanning for people already in the current scene. 
        /// Called automatically by VAM
        /// </summary>
        public override void Init()
        {
            this.LogIf(this.isDebugMode, "MorphFilterPlus.Init()");
            base.Init();
            if (!this.isInitialized)
            {
                // our own script stuff we only want to apply once
                // regardless of how many times VAM calls Init()
                this.Subscribe();
                this.ScanForUpgradableAtoms();
                this.isInitialized = true;
            }

            // the VAM-related stuff we'll apply every time Init() is called,
            // because it might call it multiple times on purpose.
            this.debugBoolJSON = new JSONStorableBool("Debug Logging", false);
            this.RegisterBool(this.debugBoolJSON);

            this.liveUpdateBoolJSON = new JSONStorableBool("Live Filtering", true);
            this.RegisterBool(this.liveUpdateBoolJSON);

            this.convertSpacesBoolJSON = new JSONStorableBool("Convert Spaces to '+'", false);
            this.RegisterBool(this.convertSpacesBoolJSON);

            this.minCharacterFloatJSON = new JSONStorableFloat("Live Filter Min Characters", 3.0f, 1f, 8f);
            this.RegisterFloat(this.minCharacterFloatJSON);

            
        }

        /// <summary>
        /// Initializes the plugin's UI, automatically called by VAM.
        /// </summary>
        public override void InitUI()
        {
            base.InitUI();

            this.CreateToggle(this.debugBoolJSON, true);
            this.CreateToggle(this.liveUpdateBoolJSON);
            this.CreateToggle(this.convertSpacesBoolJSON);
            this.CreateSlider(this.minCharacterFloatJSON);
            
        }

        /// <summary>
        /// Starts subscription to VAM's atom and scene events.
        /// these events will let us respond to loading a scene or atom.
        /// We call this on Init() and on OnEnable()
        /// </summary>
        private void Subscribe()
        {
            if (!this.isSubscribed)
            {
                SuperController.singleton.onAtomAddedHandlers += this.OnAtomAdded;
                SuperController.singleton.onAtomRemovedHandlers += this.OnAtomRemoved;
                SuperController.singleton.onSceneLoadedHandlers += this.OnSceneLoaded;
                SuperController.singleton.onSubSceneLoadedHandlers += this.OnSubSceneLoaded;

                this.isSubscribed = true;
            }
        }

        /// <summary>
        /// Ends subscription to VAM's atom and scene events.
        /// we unsubscribe on disable.
        /// </summary>
        private void Unsubscribe()
        {
            // since we unsub on disable, this gets called when we are quitting.
            // for all we know at that point the supercontroller is destroyed,
            // so we check first to avoid a null reference.
            // (We might be quitting but we can still be nice about it)
            if (SuperController.singleton != null)
            {
                SuperController.singleton.onAtomAddedHandlers -= this.OnAtomAdded;
                SuperController.singleton.onAtomRemovedHandlers -= this.OnAtomRemoved;
                SuperController.singleton.onSceneLoadedHandlers -= this.OnSceneLoaded;
                SuperController.singleton.onSubSceneLoadedHandlers -= this.OnSubSceneLoaded;                
                this.isSubscribed = false;
            }

        }


        /// <summary>
        /// Iterates the Scene's Atoms to see if there are any new people.
        /// we call this when we don't know if there are, like when
        /// becoming enabled, or when loading a new scene or subscene.        
        /// </summary>
        public void ScanForUpgradableAtoms()
        {
            if (!this.everythingIsTerrible)
            {
                this.LogIf(this.isDebugMode, "MorphFilterPlus.ScanForPeople()");

                // GetSceneAtoms() is built into VAM's MVRScript API, 
                // and fetches the scene's root objects from the Supercontroller.
                // it's nice and fast, so we iterate that for people instead of 
                // using a FindObjectOfType<> or GetComponentsInChildren<> which are slower.

                foreach (var atom in this.GetSceneAtoms())
                {
                    switch(atom.type)
                    {
                        case AtomType.Person:
                            this.UpgradePerson(atom);
                            break;                      
                    }
                    
                }
            }
        }


        /// <summary>
        /// Begins upgrading the person's input fields as a coroutine.
        /// We only make it a coroutine so we can yield a couple frames,
        /// to ensure the person is fully prepared for filter greatness.
        /// </summary>        
        public void UpgradePerson(Atom person)
        {
            // we only upgrade people that don't already have upgrades.
            // also we don't do it if stuff broke causing everything to be terrible.
            if (!this.upgradeLookup.ContainsKey(person) && !this.everythingIsTerrible)
            {
                this.LogIf(this.isDebugMode, $"MorphFilterPlus upgrading a new person named '{person.name}'");

                this.upgradeLookup[person] = new UpgradeProfile();
                this.tasks.Add(UpgradePersonProcess(person));
            }

        }


        /// <summary>
        /// This is the main process that results in 
        /// converting a standard filter input to our MorphFilterPlus input.
        /// we add it to our task list via UpgradePerson() when we find a new one.
        /// </summary>
        private IEnumerator UpgradePersonProcess(Atom person)
        {
           
            // now we start by waiting a few frames for the character to get settled in.
            // this is ugly but it's safe, so i'ma do it.
            // I don't even know if I really need to. I'm a radical.
            // radically cautious.
            yield return null;
            yield return null;
            yield return null;
            
            // ANYTHING could have happened in those 3 frames. Maybe our person's destroyed now.
            // we check that they exist and that we still care about them before continuing.
            if (person != null && this.upgradeLookup.ContainsKey(person) && !this.everythingIsTerrible)
            {                
                UpgradeProfile upgrade = this.upgradeLookup[person] as UpgradeProfile;
                
                // time to prepare for the magic. We fetch the person's male and female morph input fields,
                // using whatever named transform hierarchy path is accurate.
                Transform femaleInputField = person.transform.Find(k_inputFieldPathFemale);
                if (femaleInputField == null) femaleInputField = person.transform.Find(k_inputFieldPathFemaleAlt);

                Transform maleInputField = person.transform.Find(k_inputFieldPathMale);
                if (maleInputField == null) maleInputField = person.transform.Find(k_inputFieldPathMaleAlt);

                // schooled C# developers may note that i didn't use null propogation above.
                // schooled Unity developers will tell you that unity's Objects will LIE if you do. the Objects are LIARS

                #region create new filters to intercept input data
                // with InputFields in hand, we convert them to better ones.
                // our trick is duplicating the existing fields, disabling the old ones,
                // and having the new ones point events at us, so we intercept the input.
                // keeping the old ones around makes it easy to revert when 
                // either this plugin or the person is removed.
                #endregion

                if (femaleInputField != null && maleInputField != null)
                {

                    upgrade.OldFieldMale = maleInputField.GetComponent<InputField>();
                    upgrade.OldFieldFemale = femaleInputField.GetComponent<InputField>();

                    upgrade.NewFieldMale = this.CreateFilterField(person, upgrade.OldFieldMale);
                    upgrade.NewFieldFemale = this.CreateFilterField(person, upgrade.OldFieldFemale);

                    // the UI we are going to refresh is a sibling of the inputfield,
                    // so we find it from the parent's perspective.
                    Transform maleUITransform = upgrade.NewFieldMale.transform.parent.Find(k_morphUIPathMale);
                    Transform femaleUITransform = upgrade.NewFieldFemale.transform.parent.Find(k_morphUIPathFemale);
                    if (maleUITransform != null && femaleUITransform != null)
                    {
                        upgrade.UIMale = maleUITransform.GetComponent<GenerateDAZMorphsControlUI>();
                        upgrade.UIFemale = femaleUITransform.GetComponent<GenerateDAZMorphsControlUI>();
                        if (upgrade.UIMale == null || upgrade.UIFemale == null) this.generalError += " missing upgrade ui components !";
                    }
                    else this.generalError += " missing upgrade ui transforms !";
                }
                else
                {
                    this.generalError += " missing input fields !";
                    this.Log("Something went wrong and MorphFilterPlus couldn't find the normal input field. Did you do something?" +
                        " You changed something, didn't you? What have you done? WHAT DID YOU DO", error: true);
                }


                if (!this.hasError)
                {

                    #region get morph bank components from geometry children
                    // next, in order to actually do anything with the filter, we need to be able to
                    // access the morphs. Morphs are stored in components on child game objects of the person,
                    // one game object for each category of morph. These in turn are managed by 5 MorphBanks; 
                    // one for each gender's base and genital morphs, and an extra genital morph group for males.
                    // These 5 overarching MorphBanks are also on game objects-- they're the parents of the categories.
                    // they keep track of all of the morphs of their child subbanks.
                    // we are going to take advantage of that when we filter by using those parent morphbanks.

                    // rather than doing a GetComponent<> every time we filter, we're going to
                    // find and cache the morph banks now as part of the upgrade object.
                    #endregion

                    upgrade.MorphBanks = new DAZMorphBank[MorphFilterPlus.morphBankPaths.Length];
                    for (int i = 0; i < MorphFilterPlus.morphBankPaths.Length; i++)
                    {
                        upgrade.MorphBanks[i] = GetMorphBank(person, MorphFilterPlus.morphBankPaths[i]);
                        if (this.hasError) break;
                    }

                    upgrade.ResyncMorphs();
                    this.tasks.Add(this.DelayedAction(5, () => { upgrade.RefreshUI(); }));
                }

            }// end do upgrade
            

            if (this.hasError)
            {
                this.Log("MorphFilterPlus encountered errors: " + this.generalError, error: true);

                // if at any point we ran into a major error like a missing transform,
                // then this plugin is just not going to work, and we should give up and disable.
                this.StopWorkingAndSulk();
            }

        }


        /// <summary>
        /// Finds a child transform of the person according to a transform path,
        /// and fetches the morph bank component from it, if possible.
        /// Flags an error if it can't find the transform or morph bank.
        /// </summary>                
        private DAZMorphBank GetMorphBank(Atom person, string path)
        {
            DAZMorphBank ret_bank = null;
            Transform bankTransform = person.transform.Find(path);
            if (bankTransform != null) ret_bank = bankTransform.GetComponent<DAZMorphBank>();

            if (ret_bank == null) this.generalError += $" missing morphbank at path '{path}' !";

            return ret_bank;
        }


        /// <summary>
        /// Performs the legwork of duplicating the vanilla VAM morph search InputField 
        /// and intercepting event from the new one (and turning it blue). 
        /// flags an error if it can't find important elements.
        /// </summary>        
        private InputField CreateFilterField(Atom person, InputField oldField)
        {
            InputField ret_newField = null;
            try
            {
                ret_newField = Instantiate(oldField, oldField.transform.parent);

                // i'm not 100% sure if there's some layout stuff happening, so to preserve any
                // we make sure to plop the new field at the same spot in the hierarchy.
                ret_newField.transform.SetSiblingIndex(oldField.transform.GetSiblingIndex());

                // we don't want our duplicate field to send events to VAM's stuff,
                // so we scrub the events it might have...
                ret_newField.onValueChanged.RemoveAllListeners();
                ret_newField.onEndEdit.RemoveAllListeners();

                // ...and we add our own methods as listeners instead.
                ret_newField.onValueChanged.AddListener((value) => this.OnFilterChange(person, ret_newField));
                ret_newField.onEndEdit.AddListener((value) => this.OnFilterSubmit(person, ret_newField));

                // the clear button is duplicated too as a child
                var clearButton = ret_newField.GetComponentInChildren<Button>();
                clearButton.onClick.RemoveAllListeners();
                clearButton.onClick.AddListener(() => this.OnFilterClear(person, ret_newField));

                // add our signature MorphFilterBlue(TM)
                ret_newField.GetComponent<Image>().color = MorphFilterPlus.mintyFreshColor;

                // and we're done. wrap up by simply turning off the original input field
                oldField.gameObject.SetActive(false);
            }
            catch (Exception e)
            {
                this.generalError += "MorphFilterPlus.CreateFilterField() exception: " + e;
            }
            return ret_newField;
        }


        /// <summary>
        /// A general-use action coroutine we can add as a task
        /// to delay an arbitrary number of frames before doing something.        
        /// </summary>        
        /// <remarks>
        /// we use this to scan for people after loading a scene, for example,
        /// so that atoms are given a chance to initialize and any other plugins
        /// that might be messing with initialization are given a chance to do it.
        /// </remarks>
        private IEnumerator DelayedAction(int frameDelay, Action action)
        {
            int framesRemaining = frameDelay;
            while (!this.hasError && --framesRemaining > 0)
            {
                yield return null;
            }

            if (!this.hasError)
            {
                if (action != null) action();
            }
        }

        /// <summary>
        ///  We parse the text from the filter and apply it, IF live mode is active.         
        ///  We subscribed this to the input change event of the InputField we created
        ///  in our UpgradeField(..) method. 
        /// </summary>        
        private void OnFilterChange(Atom person, InputField field)
        {
            // we don't do anything for non-live submission mode or regex
            // and we wait for the characters to pass our threshold.
            if ((this.liveUpdateBoolJSON.val && !field.text.StartsWith("@")) && (field.text.Length >= Mathf.Floor(this.minCharacterFloatJSON.val) || field.text.Length == 0))
            {
                this.submissionTask = null;
                this.ApplyFilter(person, field);
            }

        }


        /// <summary>
        /// Clears the text of a given field, causing OnFilterChange() to be called.
        /// if live update is false, this immediately restores morphs.        
        /// We subscribed this to the new clear button's OnClick event in our UpgradeField(..) method
        /// </summary>
        private void OnFilterClear(Atom person, InputField field)
        {
            // setting the text will prompt OnFilterChange if it wasn't already empty,
            // so we don't have to do anything else here for live-mode.
            field.text = "";
            if (this.submissionTask == null && !this.hasError)
            {
                // if we haven't started to submit anything, then let's
                // restore everything and refresh the UI. If we HAVE submitted
                // something, then it's ok because we just set the field blank anyway.
                UpgradeProfile upgrade;
                if (this.upgradeLookup.TryGetValue(person, out upgrade))
                {
                    var personUpgrade = upgrade as UpgradeProfile;
                    personUpgrade.RestoreMorphs(field);
                    personUpgrade.RefreshUI(field);
                }
            }

            if (this.hasError)
            {
                this.StopWorkingAndSulk();
            }

        }


        /// <summary>
        /// Applies the inputfield as a filter IF live update is false, or we are in regex mode.        
        /// </summary>
        /// <remarks>
        /// This assigns our submission to a task we update each frame, as careful timing is necessary.
        /// Detailed comments are available within for more info.
        /// </remarks>
        private void OnFilterSubmit(Atom person, InputField field)
        {
            // we only use submission-based filtering if our live update toggle is off.
            if ((!this.liveUpdateBoolJSON.val || field.text.StartsWith("@") || field.text.Length < this.minCharacterFloatJSON.val) && !this.hasError)
            {

                #region create a carefully timed task to allow clear button events.

                // here we must be clever.
                // the submit method is called when we hit enter OR click away from the field.
                // this means it's called when we click the Clear button (because we're clicking away).
                // if we don't account for that we'll end up applying the text filter 
                // and then clearing it immediately, which will both lag and possibly break some 
                // hidden complexity within VAM's morph or UI management.

                // so what we do instead is start a submission task where we wait until a 
                // few frames after the mouse has been released before applying the text of the inputfield.
                // this gives us a chance to clear it without applying it first.                

                // this task is a cached member we replace when necessary.
                // we don't start it as a coroutine but instead step through it in update.
                #endregion

                this.submissionTask = this.SubmissionProcess(person, field);
            }            
        }

        /// <summary>
        /// The coroutine we step through to submit a filter that wasn't live updated,
        /// while ensuring we can click the clear button without double-submitting.
        /// </summary>                
        private IEnumerator SubmissionProcess(Atom person, InputField field)
        {
            // in case the mouse doesn't register as down this specific frame, give it a sec.            
            yield return null;

            // if your mouse/keyboard is stuck we don't want to loop forever, even if it is a passive loop.
            int frameLimit = 600;
            while (Input.anyKey && !this.everythingIsTerrible && --frameLimit > 0)
            {
                // we use anykey to cover all input bases even though it's probably a mouse button.
                yield return null;
            }

            if (frameLimit <= 0)
            {
                this.Log("MorphFilterPlus couldn't finish filter submission; A key or mouse was detected for 600 frames.", error: true);
                this.generalError += " submission stalled !";
            }
            else
            {
                // once the mouse is released, give it a frame so that if it was clicking
                // the clear button, it can call the clear event before we apply.
                yield return null;

                // now that no input is down and the UI did its thing, we apply.
                this.ApplyFilter(person, field);
            }
        }


        /// <summary>
        /// Fetches banks, builds the filters, and applies them to all the morphs,
        /// then refreshes the UI. This is the thing that occurs to actually
        /// make your morph ui show new stuff. It's called when an input field changes
        /// if live update is true, or after a submission event if live update is false.
        /// </summary>        
        private void ApplyFilter(Atom person, InputField field)
        {
            // this whole method structure is not very pretty or readable,
            // sort of has a function over form thing going on.
            // I'll probably prettify it later.           

            UpgradeProfile baseUpgrade;
            if (!this.hasError && this.upgradeLookup.TryGetValue(person, out baseUpgrade))
            {
                UpgradeProfile upgrade = baseUpgrade as UpgradeProfile;

                List<DAZMorphBank> banks = new List<DAZMorphBank>();

                // first we get the morph banks relevant to the field.
                if (field == upgrade.NewFieldFemale)
                {
                    banks.Add(upgrade.MorphBanks[BankIndex.FemaleChar]);
                    banks.Add(upgrade.MorphBanks[BankIndex.FemaleGen]);

                }
                else if (field == upgrade.NewFieldMale)
                {
                    banks.Add(upgrade.MorphBanks[BankIndex.MaleChar]);
                    banks.Add(upgrade.MorphBanks[BankIndex.MaleGen]);
                    banks.Add(upgrade.MorphBanks[BankIndex.MaleGen2]);
                }
                else
                {
                    this.generalError += " ApplyFilter() unrecognized field !";
                }

                if (MorphFilterPlus.ignoredMorphs == null && !this.hasError)
                {                    
                    // if this is the first time we've tried applying a filter, we'll populate our ignored morphs cache.
                    // we waited this long because some morphs seem to be generated or hidden dynamically on selection,
                    // and we would have missed them in the upgrade step. here we know for sure the character has been selected.  // update 3.0: they still sneak through!
                    MorphFilterPlus.ignoredMorphs = new HashSet<string>();
                    foreach (DAZMorphBank bank in upgrade.MorphBanks)
                    {
                        foreach (DAZMorph morph in bank.morphs)
                        {
                            // for some reason, autobehavior eyelid morphs don't get ignored in some cases.
                            // I think it's because they're generated dynamically or they start out visible and become invisible?
                            // whatever the case we can just add a manual ignore list to consider.
                            if (!morph.visible) MorphFilterPlus.ignoredMorphs.Add(morph.uid);
                        }
                    }
                    this.LogIf(this.isDebugMode, $"MorphFilterPlus is ignoring {MorphFilterPlus.ignoredMorphs.Count} morphs.");

                    // efforts to prevent eyelid morphs from appearing in the first search have led to
                    // an initial restore / refresh being called. I think it might have resolved it.
                    upgrade.RestoreMorphs();
                    upgrade.RefreshUI();

                }

                // assuming we have banks to work with, we now build our filter list to apply to morphs.
                if (string.IsNullOrEmpty(field.text.Trim()))
                {
                    // if the filter is empty, we just restore all the morphs.
                    upgrade.RestoreMorphs(field);
                }
                else if (field.text.StartsWith("@"))
                {                    
                    // new in 3.0 starting with @ uses regular expressions.
                    Regex regex = new Regex(field.text.Substring(1), RegexOptions.IgnoreCase);
                    int matchCount = 0;
                    foreach (DAZMorphBank bank in banks)
                    {
                        foreach (DAZMorph morph in bank.morphs)
                        {

                            if (MorphFilterPlus.ignoredMorphs.Contains(morph.uid)) continue;
                            try
                            {
                                morph.visible = regex.IsMatch(morph.lowerCaseResolvedDisplayName);
                            }
                            catch (Exception e)
                            {
                                this.Log($"MorphFilterPlus had an issue regex matching morph '{(morph == null ? "NULL" : morph.displayName)} : {e}", error: true);
                                morph.visible = true; // might as well show morphs we aren't sure about.
                            }
                        }
                    }                    
                }
                else
                {

                    #region the filter starts by splitting into separate requirement sets

                    // it gets tricky here but bear with me.
                    // we have two tiers of "AND" (as of 3.0), the "AND" (&) and the "PLUS" (+). the plus
                    // is treated like any other constraint but the AND and OR will create logical groups.                    
                    // AND was added to give more control of the order of operations. It is applied first.

                    // First we split the input text into multiple inputs split by the AND code.
                    // these will end up being filter sets with their own rules, and all of them will have
                    // to give the greenlight to any morph they're compared against.
                    #endregion

                    List<List<Filter>> requiredSet = new List<List<Filter>>();
                    string rawString = this.convertSpacesBoolJSON.val ? field.text.Replace(" ","+") : field.text;
                    string[] andSegments = rawString.Split(new[] { '&' }, StringSplitOptions.RemoveEmptyEntries);
                    foreach (string segment in andSegments)
                    {
                        List<Filter> optionalSet = new List<Filter>();

                        // to build a set of filters within each AND group, we first split each string into 
                        // OR segments. This will give us a final set of strings that will turn into filters.
                        // Since these are separated by OR, the morph only has to match one of them 
                        // for us to give the greenlight.
                        string[] orSegments = segment.Split(new[] { '%' }, StringSplitOptions.RemoveEmptyEntries);
                        foreach (var constraintCombo in orSegments)
                        {
                            optionalSet.Add(new Filter(constraintCombo.ToLower()));
                        }

                        requiredSet.Add(optionalSet);
                    }

                    // now that the filters are built, we are going to compare them 
                    // to each of the morph display names, and update the morph visibility.                    
                    /// using continue to avoid nesting in a few places here.
                    bool match = true;
                    foreach (DAZMorphBank bank in banks)
                    {
                        foreach (DAZMorph morph in bank.morphs)
                        {
                            
                            if (MorphFilterPlus.ignoredMorphs.Contains(morph.uid))
                            {
                                continue;
                            }

                            // by default everything is visible if we have no filters.
                            match = requiredSet.Count == 0;
                            if (match)
                            {
                                morph.visible = true;
                                continue;
                            }

                            // for the required set, we presume innocence until proven guilty.
                            match = true;
                            foreach (List<Filter> optionalSet in requiredSet)
                            {
                                if (optionalSet.Count == 0) continue;

                                // for the optional set we presume guilt until proven innocent
                                match = false;
                                foreach (Filter filter in optionalSet)
                                {
                                    match = filter.CompareString(morph.lowerCaseResolvedDisplayName);
                                    if (match) break;
                                }

                                // once we've gone through the gamut of optional filters,
                                // if we haven't matched, the required set clearly isn't all met.
                                // so we don't need to look any more.
                                if (!match) break;
                            }

                            // whatever happened, we accept the result.
                            morph.visible = match;

                        }// end morph loop
                    }// end banksy loop
                }// end 'filter cuz text' condition

                upgrade.ResyncMorphs();
                upgrade.RefreshUI(field);
                this.submissionTask = this.DelayedAction(5, upgrade.RestoreMorphs);

            }// end method valid condition


            if (this.hasError)
            {
                this.StopWorkingAndSulk();
            }

        }// end method ApplyFilters()


        /// <summary>
        /// Called by Unity every frame automatically as part of the MonoBehavior API
        /// we inherit. We use it to step through our coroutines each frame. 
        /// Update has overhead so it's a tiny bit wasteful for this. I don't want to use it.
        /// but I really REALLY don't like StartCoroutine, so I'm going to anyway for now.
        /// </summary>
        private void Update()
        {
            for (int i = 0; i < this.tasks.Count; i++)
            {
                if (!this.tasks[i].MoveNext())
                {
                    this.tasks.RemoveAt(i);
                    i--;
                }
            }

            // if we are not in live update mode,
            // then we update a submission process if we have one.
            if (this.submissionTask != null)
            {
                if (!this.submissionTask.MoveNext()) this.submissionTask = null;
            }

        }

        /// <summary>
        /// Responds to newly created atoms by checking if they're a person
        /// and upgrading them if so. This is one of our event subscriptions in OnEnable()
        /// </summary>        
        private void OnAtomAdded(Atom atom)
        {
            switch (atom.type)
            {
                case AtomType.Person:
                    this.UpgradePerson(atom);
                    break;
              
            }
        }

        /// <summary>
        /// Responds to the removal of an atom by checking if they're a person
        /// and reverting their upgrade if so. This is one of our event subscriptions in OnEnable()
        /// </summary>
        /// <param name="atom"></param>
        private void OnAtomRemoved(Atom atom)
        {
            UpgradeProfile upgrade;
            if (atom != null && this.upgradeLookup.TryGetValue(atom, out upgrade))
            {
                this.LogIf(this.isDebugMode, "MorphFilterPlus Reverting mods for removed atom: " + atom.name);
                upgrade?.Revert();
                this.upgradeLookup.Remove(atom);
            }

        }

        /// <summary>
        /// Responds to the VAM scene being loaded by scanning for people present,
        /// after a brief delay to allow any weird atom things to be settled.
        /// </summary>
        private void OnSceneLoaded()
        {
            IEnumerator delayedScan = this.DelayedAction(frameDelay: 5, action: this.ScanForUpgradableAtoms);
            this.tasks.Add(delayedScan);
        }


        /// <summary>
        /// Responds to a VAM subscene being loaded by scanning for new people,
        /// after a brief delay to allow any weird atom things to be settled.
        /// </summary>        
        private void OnSubSceneLoaded(SubScene scene)
        {
            IEnumerator delayedScan = this.DelayedAction(frameDelay: 5, action: this.ScanForUpgradableAtoms);
            this.tasks.Add(delayedScan);
        }


        /// <summary>
        /// Sets the flag implicating that something is critically wrong,
        /// and disables the script. This is called when we encounter an issue
        /// that would prevent the plugin from working, such as missing transforms.
        /// Since it disables itself, it results in the reversion of any existing upgrades.
        /// </summary>
        private void StopWorkingAndSulk()
        {
            if (!this.everythingIsTerrible)
            {

                this.Log($"MorphFilterPlus is stopping due to errors: {this.generalError} \nThings didn't go as planned, and MorphFilterPlus has stopped working for now. " +
                    "You can try restarting or reloading it if you want.", error: true);

                this.everythingIsTerrible = true;
                this.generalError = "";
            
                if (this.enabledJSON != null && this.enabledJSON.toggle != null)
                {
                    this.enabledJSON.toggle.isOn = false;
                }

            }
        }



        private void LogIf(bool condition, string str, bool error = false)
        {
            if (condition) this.Log(str);
        }

        private void Log(string str, bool error = false)
        {
            if (error) SuperController.LogError(str);
            else SuperController.LogMessage(str);
        }
       

        /// <summary>
        /// Handles references and effects for a specific person and their UI elements,
        /// Persons have morph input fields, ui, and banks we track in addidion to the 
        /// base save appearance button.
        /// </summary>
        private class UpgradeProfile
        {            
            
            public InputField OldFieldFemale;
            public InputField OldFieldMale;

            public InputField NewFieldFemale;
            public InputField NewFieldMale;

            public GenerateDAZMorphsControlUI UIFemale;
            public GenerateDAZMorphsControlUI UIMale;

            public DAZMorphBank[] MorphBanks;
         

            /// <summary>
            /// Unhides all morphs for this person except ignored ones.
            /// called when removing the plugin or person, as well as 
            /// before saving to ensure the morphs can be saved.
            /// </summary>
            public void RestoreMorphs()
            {                
                if (this.MorphBanks != null && MorphFilterPlus.ignoredMorphs != null)
                {
                    foreach (var bank in this.MorphBanks)
                    {
                        this.RestoreBank(bank);
                    }
                }
                this.ResyncMorphs();
            }
            

            /// <summary>
            /// Restores morphs only relevant to the input field (fem or male)
            /// </summary>            
            public void RestoreMorphs(InputField field)
            {
                if (this.MorphBanks != null && MorphFilterPlus.ignoredMorphs != null)
                {                    
                    if (field == this.NewFieldFemale)
                    {
                        this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.FemaleChar]);
                        this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.FemaleGen]);
                    }
                    else if (field == this.NewFieldMale)
                    {
                        this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.MaleChar]);
                        this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.MaleGen]);
                        this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.MaleGen2]);
                    }
                    this.ResyncMorphs();
                }                
            }

            /// <summary>
            /// iterates all morphs in a bank to restore them to their pristine visibility.
            /// </summary>
            private void RestoreBank(DAZMorphBank bank)
            {
                if (bank != null)
                {
                    foreach (var morph in bank.morphs)
                    {
                        morph.visible = !MorphFilterPlus.ignoredMorphs.Contains(morph.uid);
                    }
                }
            }

            /// <summary>
            /// Ensures the UI for the person has the most up-to-date morph settings,
            /// so that they can be written or re-displayed.
            /// </summary>
            public void ResyncMorphs()
            {
                if (this.UIMale != null) this.UIMale.ResyncMorphs();
                if (this.UIFemale != null) this.UIFemale.ResyncMorphs();
            }

            /// <summary>
            /// Refreshes the drawn morphs on any active morph page.
            /// </summary>
            public void RefreshUI()
            {

                if (this.UIMale != null && this.UIMale.gameObject.activeInHierarchy) this.UIMale.ForceCategoryRefresh();
                if (this.UIFemale != null && this.UIFemale.gameObject.activeInHierarchy) this.UIFemale.ForceCategoryRefresh();
            }

            /// <summary>
            /// refreshes the drawn morphs on the page associated with a search bar,
            /// if the page is active.
            /// </summary>
            public void RefreshUI(InputField field)
            {
                if (field == this.NewFieldFemale && this.UIFemale != null && this.UIFemale.gameObject.activeInHierarchy)
                {                    
                    this.UIFemale.ForceCategoryRefresh();
                }
                else if (field == this.NewFieldMale && this.UIMale != null && this.UIMale.gameObject.activeInHierarchy)
                {                    
                    this.UIMale.ForceCategoryRefresh();
                }
            }


            /// <summary>
            /// Destroys all generated objects, removes all modifications,
            /// unsubscribes from everything and dumps all references,
            /// in a complete reversion of the upgrade. Called when the person 
            /// or plugin is removed.
            /// </summary>
            public void Revert()
            {                
                this.RestoreMorphs();
                this.RefreshUI();
                if (this.OldFieldFemale != null) this.OldFieldFemale.gameObject.SetActive(true);
                if (this.OldFieldMale != null) this.OldFieldMale.gameObject.SetActive(true);

                if (this.NewFieldFemale != null) Destroy(this.NewFieldFemale.gameObject);
                if (this.NewFieldMale != null) Destroy(this.NewFieldMale.gameObject);

                // while not strictly necessary, we want to make sure references are released
                // to avoid memory bloat, in case VAM or Unity or a plugin maintains a reference
                // to this object somewhere, somehow.
                this.OldFieldFemale = null;
                this.OldFieldMale = null;

                this.UIMale = null;
                this.UIFemale = null;
                this.MorphBanks = null;

            }


        }

    }
}